#ifndef FERI_RICKER
#define FERI_RICKER
// -----------------------------------------------
// ATTRACTOR ANALYSIS WITH GRAPH REPRESENTATION
//
// Ricker Map

// -----------------------------------------------

// CHECK STRIPE UPON USING SOURCES AND SINKS CAN BE USEFUL
//#define FERI_CHECK_STRIPE

// FINDING ATTRACTION DOMAIN WITH CONSTRUCTING LYAPUNOV FUNCTION (SEEN FROM DELANOUE)
//#define FERI_LYAPUNOV

// BOXES 45 DEG ROTATED
//#define FERI_ROTATED

// MIDPOINT EVALUATION
#define FERI_MIDPOINT

// -----------------------------------------------

#include "main.hpp"
#include "typedefs.hpp"

// -----------------------------------------------

// 2D Ricker Map
template<typename T, int D>
class VectorField : public capd::dynsys::FadMap<T, D>
{
	public:

		typedef typename capd::dynsys::FadMap<T, D> BaseClass;
		typedef typename BaseClass::MatrixType MatrixType;
		typedef typename BaseClass::VectorType VectorType;
		typedef typename BaseClass::TFVector TFVector;


		// constructor
		VectorField ()
		{}

		VectorField (const T &ALPHA)
		{
			PARAM_a = ALPHA;
		}

		// -----------------------------------------------

		// FadBad::vectorfield
		template <typename AVector>
		AVector operator()(const AVector &in) const
		{
			AVector out(2);

			#ifdef FERI_ROTATED
				out(1) = (in(1) + in(2)) * (exp(PARAM_a - in(1) + in(2)) + 1) / interval(2.0);
				out(2) = (in(1) + in(2)) * (exp(PARAM_a - in(1) + in(2)) - 1) / interval(2.0);
			#else
				// 1 iterate
				//out[0] = in[1];
				//out[1] = in[1] * exp(PARAM_a - in[0]);

				// 2 iterates
				//out[0] = in[1] * exp(PARAM_a - in[0]);
				//out[1] = in[1] * exp(2 * PARAM_a - in[0] - in[1]);

				// 3 iterates
				out[0] = in[1] * exp(2 * PARAM_a - in[0] - in[1]);
				out[1] = in[1] * exp(3 * PARAM_a - in[0] - in[1] * (1 + exp(PARAM_a - in[0])));

				// 4 iterates
				//out[0] = in[1] * exp(3 * PARAM_a - in[0] - in[1] * (1 + exp(PARAM_a - in[0])));
				//out[1] = in[1] * exp(4 * PARAM_a - in[0] - in[1] * (1 + exp(PARAM_a - in[0]) + exp(2 * PARAM_a - in[0] - in[1])));
			#endif


			return out;
		}

		// FadBad::derivative
		MatrixType operator[](const VectorType& u) const
		{
			MatrixType der(2, 2);
			computeDerivative(*this, u, der);
			return der;
		}

		// FadBad::returns the dimension
		int getDimension() const
		{
			return D;
		}

		// parameter of the ricker map
		T PARAM_a;

		// evaluate the vectorfield
		void evaluate(const VectorType &in, VectorType &out);

		// constructing the starting invariant set
		void constructInvariantSet(VectorType &V);

		// constructing the local domain of attraction
		void findAttractionDomain(VectorType &attDomain);

		// resetting graphics
		void resetGraphics(const double &ScreenSize);

		// plotting the boxes
		void plotFixedPoint(const double &SCREENSIZE);

		// deleting the edges connecting the left and right side of the stripe
		bool checkRightStripe(const double &leftStripe, GraphType &GRAPH);

		// check if the vertices left from the stripe have inEdges only from themselves
		bool checkLeftStripe(const double &leftStripe, const GraphType &GRAPH);

		// checking if on the left side of the stripe we have separation and if yes separates in the graph
		void checkStripe(const double &leftStripe, GraphType &GRAPH, bool &leftSeparated, bool &rightEmpty);

		// constructing the local domain of attraction
		bool checkHessian(const interval &x, const interval &y);

		// constructing the local domain of attraction
		void findLyapAttractionDomain(const VectorType &V, VectorType &attDomain);

		// adding initial boxes (added here as well, because of vectorfield-specific option for rotation)
		void buildRotatedInitialBoxes(const VectorType &V, const double &initialBoxSize, GraphType &GRAPH);

		// check attraction to fixed point, true if all boxes are attracted
		bool checkRotatedAttractionDomain(const VectorType &attDomain, GraphType &GRAPH);

	protected:
};

// -----------------------------------------------

// evaluate Function
template<typename T, int D>
void VectorField<T, D>::evaluate(const VectorType &in, VectorType &out)
{
	#ifdef FERI_MIDPOINT
		VectorType in_x;

		in_x = midVector(in);
		out = (*this)(in_x) + (*this)[in] * (in - in_x);
	#else
		out = (*this)(in);
	#endif
}

// constructing the invariant set
template<typename T, int D>
void VectorField<T, D>::constructInvariantSet(VectorType &V)
{
	long int iter;
	interval hi, gi, hi1;

	// we try to refine invariant sets until the change is less than this
	const double inv_TOL = 1e-9;

	// max iterations if we don't reach other stopping condition
	const long int inv_maxITER = 1000000;



	iter = 0;

	hi = interval(-1, PARAM_a.rightBound());
	hi1 = interval(0, PARAM_a.rightBound());

	// iterate until H is not changed OR the change in the left bound of H is less then tolerance OR we reached the num of max iterations
	while ( (hi1 != hi) && ( ( hi1.leftBound() - hi.leftBound() > inv_TOL ) || (iter < inv_maxITER) ) ) {
		++iter;
		hi = hi1;
		// G = [ a, a * e^(2a-2H) ]
		// H = [ a * e^(2a-2G), a ]
		gi  = interval( PARAM_a.leftBound(), (PARAM_a * exp(2.0 * PARAM_a - 2.0 * hi.left())).rightBound() );
		hi1 = interval( (PARAM_a * exp(2.0 * PARAM_a - 2.0 * gi.right())).leftBound(), PARAM_a.rightBound());
	}

	hi = hi1;
	gi = interval(PARAM_a.leftBound(), (PARAM_a * exp(2.0 * PARAM_a - 2.0 * hi.left())).rightBound() );

	V(1) = intervalHull(hi, gi);
	V(2) = V(1);
}

// LOCAL ATTRACION DOMAIN: METHOD - EIGENBASE
template<typename T, int D>
void VectorField<T, D>::findAttractionDomain(VectorType &attDomain)
{
	interval epsilon, epsilon2;

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif

	if ( PARAM_a.rightBound() < 1 ) {	
	    epsilon = (capd::intervals::power( interval(1.0) - capd::intervals::sqrt(PARAM_a), 2 ) * capd::intervals::sqrt(4.0 * PARAM_a - 1) / (1 - capd::intervals::power(PARAM_a, 2)) - 1e-16).leftBound();
	} else {
	    epsilon = 0;
	}
	if ( PARAM_a.leftBound() > (interval(999) / interval(1000)).rightBound() ) {
		epsilon2 = interval(1.0) / interval(22.0);
	} else if ( PARAM_a.leftBound() > (interval(875) / interval(1000)).rightBound() ) {
		epsilon2 = interval(1.0) / interval(37.0);
	} else {
		epsilon2 = epsilon;
	}
	epsilon = capd::max(epsilon.leftBound(), epsilon2.leftBound());

	attDomain(1) = interval( (PARAM_a.right() - epsilon).leftBound(), (PARAM_a.left() + epsilon).rightBound() );
	attDomain(2) = attDomain(1);

	#ifdef FERI_GRAPHICS_DRAW_EXTRA

		#ifdef FERI_ROTATED
			double center, epsi;

			epsi = epsilon.leftBound();
			epsi = epsi / sqrt(2.0);
			center = (attDomain(1).rightBound() + attDomain(1).leftBound()) / 2.0;

			lt1 = center - epsi;
			lt2 = - epsi;
			rb1 = center + epsi;
			rb2 = epsi;
		#else
			lt1 = attDomain(1).leftBound();
			lt2 = attDomain(2).leftBound();
			rb1 = attDomain(1).rightBound();
			rb2 = attDomain(2).rightBound();
		#endif

		rootFrame.box(lt1, lt2, rb1, rb2, YELLOW);
		waitBt();
	#endif
}

// -----------------------------------------------

// resetting graphics
template<typename T, int D>
void VectorField<T, D>::resetGraphics(const double &ScreenSize)
{
	rootFrame.Clear();
	#ifdef FERI_ROTATED
		rootFrame.line(0., 0., ScreenSize, 0.);
	#else
		rootFrame.line(0., 0., ScreenSize, ScreenSize);
	#endif
}

// plotting the boxes
template<typename T, int D>
void VectorField<T, D>::plotFixedPoint(const double &SCREENSIZE)
{
	double lt1, lt2, rb1, rb2;
	interval fpoint;


	fpoint = PARAM_a.mid() + interval(-1,1) * SCREENSIZE / 100.0;

	lt1 = fpoint.leftBound();
	rb1 = fpoint.rightBound();

	#ifdef FERI_ROTATED
		lt2 = - SCREENSIZE / 100.0;
		rb2 = SCREENSIZE / 100.0;
	#else
		lt2 = lt1;
		rb2 = rb1;
	#endif

	rootFrame.boxFill(lt1, lt2, rb1, rb2, RED);
	waitBt();
}

// -----------------------------------------------

// deleting the edges connecting the left and right side of the stripe
template<typename T, int D>
bool VectorField<T, D>::checkRightStripe(const double &leftStripe, GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	VertexDescriptor vertex;

	std::pair<inEdgeIterator, inEdgeIterator> inEdges;
	inEdgeIterator ine, ine2;
	bool result = true;

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif

	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;


	while ( it != GRAPHlimits.second ) {

		if ( GRAPH[*it].encl(1).rightBound() > leftStripe ) {

			inEdges = boost::in_edges(*it, GRAPH);
			ine = inEdges.first;

			while ( ine != inEdges.second ) {
				vertex = source(*ine, GRAPH);
				if ( GRAPH[vertex].encl(1).rightBound() <= leftStripe ) {
					ine2 = ine;
					++ine;
					boost::remove_edge(*ine2, GRAPH);
					result = false;

					#ifdef FERI_GRAPHICS_DRAW_EXTRA
						lt1 = GRAPH[vertex].encl(1).leftBound();
						lt2 = GRAPH[vertex].encl(2).leftBound();
						rb1 = GRAPH[vertex].encl(1).rightBound();
						rb2 = GRAPH[vertex].encl(2).rightBound();

						rootFrame.boxFill(lt1, lt2, rb1, rb2, RED);

						lt1 = GRAPH[*it].encl(1).leftBound();
						lt2 = GRAPH[*it].encl(2).leftBound();
						rb1 = GRAPH[*it].encl(1).rightBound();
						rb2 = GRAPH[*it].encl(2).rightBound();

						rootFrame.boxFill(lt1, lt2, rb1, rb2, GREEN);
					#endif

				} else {
					++ine;
				}
			}

		}
	++it;
	}
	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		waitBt();
	#endif
	return result;


}

// check if the vertices left from the stripe have inEdges only from themselves
template<typename T, int D>
bool VectorField<T, D>::checkLeftStripe(const double &leftStripe, const GraphType &GRAPH)
{
	bool result = true;

	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	VertexDescriptor vertex;

	std::pair<inEdgeIterator, inEdgeIterator> inEdges;
	inEdgeIterator ine;

	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;


	while ( it != GRAPHlimits.second ) {

		if ( GRAPH[*it].encl(1).rightBound() <= leftStripe ) {

			inEdges = boost::in_edges(*it, GRAPH);
			ine = inEdges.first;

			while ( ine != inEdges.second ) {
				vertex = source(*ine, GRAPH);
				if ( GRAPH[vertex].encl(1).rightBound() > leftStripe ) {
					result = false;
					break;
				}
				++ine;
			}

		}
		++it;

	}

	return result;
}

// checking if on the left side of the stripe we have separation and if yes separates in the graph
template<typename T, int D>
void VectorField<T, D>::checkStripe(const double &leftStripe, GraphType &GRAPH, bool &leftSeparated, bool &rightEmpty)
{
	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		#ifdef FERI_ROTATED
			rootFrame.line(leftStripe, - PARAM_a.rightBound(), leftStripe, PARAM_a.rightBound(), RED);
		#else
			rootFrame.line(leftStripe, 0., leftStripe, PARAM_a.rightBound(), RED);
		#endif
	#endif

	if ( not(leftSeparated) ) {
		leftSeparated = checkLeftStripe(leftStripe, GRAPH);
	} else if ( not(rightEmpty) ) {
		rightEmpty = checkRightStripe(leftStripe, GRAPH);
	}
}

// -----------------------------------------------

// LOCAL ATTRACION DOMAIN CONSTRUCTION: METHOD - CONSTRUCTING LYAPUNOV FUNCTION
// checking the Hessian
template<typename T, int D>
bool VectorField<T, D>::checkHessian(const interval &x, const interval &y)
{
	MatrixType Hessian, VHessian;
	bool result = true;
	int lr[4];
	T a = PARAM_a;


	Hessian(1, 1) = -(4+2*power(a,2)+2*a)/(-2+power(a,2)+a)-(2*y-2*a)*y*exp(a-x)/(-2+power(a,2)+a)-y*exp(a-x)*((2*y-2*a)/(-2+power(a,2)+a)-(2+2*a)*(y*exp(a-x)-a)/(a*(-2+power(a,2)+a)))+2*power(y,2)*power((exp(a-x)),2)*(2+2*a)/(a*(-2+power(a,2)+a))+(y*exp(a-x)-a)*(2+2*a)*y*exp(a-x)/(a*(-2+power(a,2)+a));
	Hessian(1, 2) = 4/(-2+power(a,2)+a)+2*y*exp(a-x)/(-2+power(a,2)+a)+(2*y-2*a)*exp(a-x)/(-2+power(a,2)+a)+exp(a-x)*((2*y-2*a)/(-2+power(a,2)+a)-(2+2*a)*(y*exp(a-x)-a)/(a*(-2+power(a,2)+a)))+y*exp(a-x)*(2/(-2+power(a,2)+a)-(2+2*a)*exp(a-x)/(a*(-2+power(a,2)+a)))-power((exp(a-x)),2)*(2+2*a)*y/(a*(-2+power(a,2)+a))-(y*exp(a-x)-a)*(2+2*a)*exp(a-x)/(a*(-2+power(a,2)+a));
	Hessian(2, 1) = Hessian(1, 2);
	Hessian(2, 2) = -(4+4*a)/(a*(-2+power(a,2)+a))+(4+2*power(a,2)+2*a)/(-2+power(a,2)+a)-4*exp(a-x)/(-2+power(a,2)+a)-2*exp(a-x)*(2/(-2+power(a,2)+a)-(2+2*a)*exp(a-x)/(a*(-2+power(a,2)+a)));



	for (lr[0] = 1; lr[0] <= 2; ++lr[0])
		for (lr[1] = 1; lr[1] <= 2; ++lr[1])
			for (lr[2] = 1; lr[2] <= 2; ++lr[2])
				for (lr[3] = 1; lr[3] <= 2; ++lr[3]) {
					(lr[0] == 1 ? VHessian(1, 1) = Hessian(1, 1).left() : VHessian(1, 1) = Hessian(1, 1).right());
					(lr[1] == 1 ? VHessian(1, 2) = Hessian(1, 2).left() : VHessian(1, 2) = Hessian(1, 2).right());
					(lr[2] == 1 ? VHessian(2, 1) = Hessian(2, 1).left() : VHessian(2, 1) = Hessian(2, 1).right());
					(lr[3] == 1 ? VHessian(2, 2) = Hessian(2, 2).left() : VHessian(2, 2) = Hessian(2, 2).right());

					result = result && ((VHessian(1, 1) > 0) && (VHessian(1, 1) * VHessian(2, 2) - VHessian(2, 1) * VHessian(1, 2) > 0));
				}

	return result;
}

// constructing the local domain of attraction
template<typename T, int D>
void VectorField<T, D>::findLyapAttractionDomain(const VectorType &V, VectorType &attDomain)
{
	const double shrinkingFactor = 0.9;

	bool go_on = true;
	double rx, ry;

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif



	rx = ( capd::min(PARAM_a.left() - V(1).left(), V(1).right() - PARAM_a.right()) ).leftBound();
	ry = ( capd::min(PARAM_a.left() - V(2).left(), V(2).right() - PARAM_a.right()) ).leftBound();

	while (go_on) {

		rx = rx * shrinkingFactor;
		ry = ry * shrinkingFactor;

		attDomain(1) = PARAM_a + rx * interval(-1, 1);
		attDomain(2) = PARAM_a + ry * interval(-1, 1);

		#ifdef FERI_GRAPHICS_DRAW_EXTRA

			#ifdef FERI_ROTATED
				double epsilon, center;

				epsilon = (attDomain(1).rightBound() - attDomain(1).leftBound()) / 2.0 / sqrt(2.0);
				center = (attDomain(1).rightBound() + attDomain(1).leftBound()) / 2.0;

				lt1 = center - epsilon;
				lt2 = - epsilon;
				rb1 = center + epsilon;
				rb2 = epsilon;
			#else
				lt1 = attDomain(1).leftBound();
				lt2 = attDomain(2).leftBound();
				rb1 = attDomain(1).rightBound();
				rb2 = attDomain(2).rightBound();
			#endif

			rootFrame.box(lt1, lt2, rb1, rb2, YELLOW);
		#endif

		go_on = not( checkHessian(attDomain(1), attDomain(2)) );

	}

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		rootFrame.boxFill(lt1, lt2, rb1, rb2, GREEN);
		waitBt();
	#endif

}
// END LOCAL ATTRACION DOMAIN CONSTRUCTION

// -----------------------------------------------

// adding initial boxes (added here as well, because of vectorfield-specific option for rotation)
template<typename T, int D>
void VectorField<T, D>::buildRotatedInitialBoxes(const VectorType &V, const double &initialBoxSize, GraphType &GRAPH)
{
	double x1, y1, x2, y2;
	VertexType vbox;

	vbox.visited = false;

	x1 = V(1).leftBound();
	x2 = x1 + initialBoxSize;

	while ( x1 < V(1).rightBound() ) {

		y1 = 0;
		y2 = initialBoxSize;

		while (( y1 < (V(2).rightBound() - interval(x1, x2)) ) && ( y1 < interval(x1,x2) )) {

			vbox.encl(1) = interval(x1, x2);
			vbox.encl(2) = interval(y1, y2);
			boost::add_vertex(vbox, GRAPH);

			vbox.encl(2) = interval(- y2, - y1);
			boost::add_vertex(vbox, GRAPH);

			y1 = y2;
			y2 = capd::min(y2 + initialBoxSize, (V(2) / interval(2.0)).rightBound());

		}

		x1 = x2;
		x2 = capd::min(x2 + initialBoxSize, V(1).rightBound());

	}
}

// check attraction to fixed point, true if all boxes are attracted (Rotated Coordinates)
template<typename T, int D>
bool VectorField<T, D>::checkRotatedAttractionDomain(const VectorType &attDomain, GraphType &GRAPH)
{
	std::pair<VertexIterator, VertexIterator> GRAPHlimits;
	VertexIterator it, it2;

	VectorType box, tbox;

	bool result = true;

	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif

	#ifdef FERI_GRAPHICS_DRAW_EXTRA

		double epsilon, center;

		epsilon = (attDomain(1).rightBound() - attDomain(1).leftBound()) / 2.0 / sqrt(2.0);
		center = (attDomain(1).rightBound() + attDomain(1).leftBound()) / 2.0;

		lt1 = center - epsilon;
		lt2 = - epsilon;
		rb1 = center + epsilon;
		rb2 = epsilon;

		rootFrame.box(lt1, lt2, rb1, rb2, YELLOW);
	#endif



	GRAPHlimits = boost::vertices(GRAPH);
	it = GRAPHlimits.first;

	while ( it != GRAPHlimits.second ) {

		box = GRAPH[*it].encl;

		tbox(1) = box(1) - box(2);
		tbox(2) = box(1) + box(2);

		if ( subset(tbox, attDomain) ) {

			#ifdef FERI_GRAPHICS_DRAW_EXTRA
				lt1 = box(1).leftBound();
				lt2 = box(2).leftBound();
				rb1 = box(1).rightBound();
				rb2 = box(2).rightBound();

				rootFrame.boxFill(lt1, lt2, rb1, rb2, GREEN);
			#endif

			it2 = it;
			++it;
			boost::clear_vertex(*it2, GRAPH);
			boost::remove_vertex(*it2, GRAPH);

		} else {

			result = false;
			++it;

		}

	}

	#ifdef FERI_GRAPHICS
	waitBt();
	#endif

	return result;
}

// ~ end of file ~
#endif
